#!/usr/bin/env python

import sys
import subprocess
import logging
import argparse
from os.path import join as pjoin
import divvy


import pandas as pd
from imcpipeline import LOGGER as log


def main(cli=None) -> int:
    log.info("IMCpipeline runner")
    parser = parse_arguments()
    args, unknown = parser.parse_known_args(cli)
    # the extra arguments will be passed to the pipeline and
    # compounded arguments (mostly the --cellprofiler-exec argument)
    # should be quoted again
    args.cli = ["'" + x + "'" if " " in x else x for x in unknown]

    log.info("Generating project from given CSV annotation.")
    annot = pd.read_csv(args.metadata).set_index(args.sample_file_attribute)
    if args.toggle:
        log.info("Subsampling samples based on the `toggle` column.")
        annot = annot.loc[annot["toggle"].isin([1, "1", True, "TRUE", "True"]), :]

    log.info("Setting compute settings using divvy.")
    compute = divvy.ComputingConfiguration()
    compute.activate_package(args.compute)

    # Now prepare job submission
    jobs = list()
    cli_args = " ".join(args.cli)
    # the '--' is to separate the nargs from the positional in case there aren't more args
    if cli_args == "":
        cli_args = "--"
    for sample, _ in annot.iterrows():
        log.info("Processing sample %s", sample)

        input_dir = pjoin(args.input_dir, sample)
        output_dir = pjoin(args.output_dir, sample)

        cmd = f"imcpipeline {cli_args} -i {input_dir} -o {output_dir}"

        job_name = f"imcpipeline_{sample}"
        output_prefix = pjoin("submission", job_name)
        job_file = output_prefix + ".sh"
        data = {
            "jobname": job_name,
            "logfile": output_prefix + ".log",
            "mem": args.mem,
            "cores": args.cores,
            "time": args.time,
            "partition": args.partition,
            "code": cmd,
        }

        compute.write_script(job_file, data)
        jobs.append(job_file)

    log.info("Submitting jobs.")
    cmd = compute.get_active_package()["submission_command"]

    if not args.dry_run:
        for job in jobs:
            print(cmd, job)
            subprocess.call([cmd, job])

    log.info("Finished with all samples.")
    return 0


def parse_arguments() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    msg = "The corresponding attribute to be passed to the job scheduler."
    parser.add_argument("--mem", dest="mem", default="48G", help=msg)
    parser.add_argument("--cores", dest="cores", default=4, help=msg)
    parser.add_argument("--time", dest="time", default="02:00:00", help=msg)
    parser.add_argument("--partition", dest="partition", default="panda", help=msg)
    choices = divvy.ComputingConfiguration().list_compute_packages()
    msg = "`Divvy` compute configuration to be used when submitting the jobs."
    parser.add_argument("--divvy-configuration", dest="compute", choices=choices, help=msg)
    msg = "Whether to do all steps except job submission."
    parser.add_argument(
        "-d", "--dry-run", dest="dry_run", action="store_true", default=False, help=msg,
    )
    msg = "Attribute in sample annotation containing the path to the input files."
    parser.add_argument(
        "--attribute", dest="sample_file_attribute", default="sample_name", help=msg,
    )
    msg = "The parent directory of containting the input data."
    parser.add_argument("--input-dir", dest="input_dir", default="data", help=msg)
    msg = "Parent directory for output files."
    parser.add_argument("--output-dir", dest="output_dir", default="processed", help=msg)
    msg = "CSV file with metadata for all samples."
    parser.add_argument(dest="metadata", help=msg)
    msg = (
        "Whether all samples or only samples marked with a positive value in the `toggle`"
        "column should be processed."
    )
    parser.add_argument("--toggle", dest="toggle", action="store_true", default=False, help=msg)

    return parser


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(1)



#!/usr/bin/env python

"""
Utilities for the imcpipeline.
"""

import sys
import textwrap
import tarfile
import os
import resource
from os.path import join as pjoin
import tempfile
import subprocess
import re
import shutil
import urllib.request as request
import requests
from contextlib import closing
from typing import Callable

import pkg_resources

from imcpipeline import (
    LOGGER as log,
    DOCKER_IMAGE,
    DEMO_DATA,
    DEMO_CSV,
    DEMO_ILASTIK_MODEL,
)
import imcpipeline.config as cfg


def download_file(url: str, output_file: str, chunk_size=1024) -> None:
    """
    Download a file and write to disk in chunks (not in memory).

    Parameters
    ----------
    url : :obj:`str`
        URL to download from.
    output_file : :obj:`str`
        Path to file as output.
    chunk_size : :obj:`int`
        Size in bytes of chunk to write to disk at a time.
    """
    if url.startswith("ftp://"):

        with closing(request.urlopen(url)) as r:
            with open(output_file, "wb") as f:
                shutil.copyfileobj(r, f)
    else:
        response = requests.get(url, stream=True)
        with open(output_file, "wb") as outfile:
            outfile.writelines(response.iter_content(chunk_size=chunk_size))


def docker_or_singularity() -> str:
    for runner in ["docker", "singularity"]:
        if shutil.which(runner):
            log.info("Selecting '%s' as container runner.", runner)
            return runner
    raise ValueError("Neither docker or singularity are available!")


def check_requirements(func: Callable) -> Callable:
    """
    Decorator to make sure pipeline requirements are met before running a function.
    """

    def inner():
        if cfg.args.containerized is not None:
            if cfg.args.containerized == "singularity":
                cfg.args.container_image = (
                    "docker://" + cfg.args.container_image
                )
        if cfg.args.cellprofiler_plugin_path is None:
            get_zanotelli_code("cellprofiler_plugin_path", "ImcPluginsCP")
        if cfg.args.cellprofiler_pipeline_path is None:
            get_zanotelli_code(
                "cellprofiler_pipeline_path", "ImcSegmentationPipeline"
            )
        func()

    return inner


def get_zanotelli_code(arg: str, repo: str) -> None:
    """Download CellProfiler plugins from Zanotelli et al."""
    if repo not in ["ImcSegmentationPipeline", "ImcPluginsCP"]:
        raise ValueError("Please choose only one of the two available repos.")
    _dir = os.path.abspath(
        pjoin(os.path.curdir, cfg.args.external_lib_dir, repo)
    )
    if not os.path.exists(_dir):
        os.makedirs(os.path.dirname(_dir), exist_ok=True)
        url = f"https://github.com/BodenmillerGroup/{repo} {_dir}"
        cmd = f"git clone {url}"
        run_shell_command(cmd)
    setattr(cfg.args, arg, _dir)


def build_docker_image():
    """Build docker image that includes the CellProfiler plugins by Zanotelli et al."""
    tmpdir = tempfile.TemporaryDirectory()
    os.makedirs(pjoin(tmpdir.name, "docker"), exist_ok=True)

    dockerfile = pkg_resources.resource_filename(
        "imcpipeline", "docker/Dockerfile"
    )
    shutil.copy(dockerfile, pjoin(tmpdir.name, "docker", "Dockerfile"))

    cmd = f"docker -t {DOCKER_IMAGE} {pjoin(tmpdir.name, 'docker')}"
    run_shell_command(cmd)


def get_docker_image_or_build() -> None:
    """Pull docker image to use or build it if not existing."""

    def check_image() -> bool:
        try:
            # check if exists
            out = (
                subprocess.check_output("docker images".split(" "))
                .decode()
                .strip()
            )
            for line in out.split("\n")[1:]:
                if line.split(" ")[0] == DOCKER_IMAGE:
                    return True
        except FileNotFoundError:
            log.error("Docker installation not detected.")
            raise
        except IndexError:
            pass
        return False

    if not check_image():
        log.debug("Did not find cellprofiler docker image. Will build.")
        build_docker_image()
    else:
        log.debug("Found docker image.")
    cfg.args.container_image = DOCKER_IMAGE


def check_ilastik(func: Callable) -> Callable:
    """A decorator to check whether ilastik executable exists and if not get it."""

    def get_ilastik(version="1.3.3post2"):
        url = "https://files.ilastik.org/"
        file = f"ilastik-{version}-Linux.tar.bz2"
        os.makedirs(cfg.args.external_lib_dir, exist_ok=True)
        log.info("Downloading ilastik archive.")
        download_file(url + file, pjoin(cfg.args.external_lib_dir, file))
        # download_file(url + file + ".sha256", pjoin(cfg.args.external_lib_dir, file) + ".sha256")
        # f"sha256sum -c {url + file + '.sha256'}"
        log.info("Extracting ilastik archive.")
        tar = tarfile.open(pjoin(cfg.args.external_lib_dir, file), "r:bz2")
        tar.extractall(cfg.args.external_lib_dir)
        tar.close()

    def inner():
        def_ilastik_sh_path = pjoin(
            cfg.args.external_lib_dir,
            "ilastik-1.3.3post2-Linux",
            "run_ilastik.sh",
        )
        if cfg.args.ilastik_sh_path is None:
            if not os.path.exists(def_ilastik_sh_path):
                os.makedirs(cfg.args.external_lib_dir, exist_ok=True)
                get_ilastik()
            cfg.args.ilastik_sh_path = def_ilastik_sh_path
        func()

    return inner


def run_shell_command(cmd) -> int:
    """
    Run a system command.

    Will detect whether a separate shell is required.
    """
    # in case the command has unix pipes or bash builtins,
    # the subprocess call must have its own shell
    # this should only occur if cellprofiler is being run uncontainerized
    # and needs a command to be called prior such as conda activate, etc
    symbol = any([x in cmd for x in ["&", "&&", "|"]])
    source = cmd.startswith("source")
    shell = bool(symbol or source)
    log.debug(
        "Running command%s:\n%s",
        " in shell" if shell else "",
        textwrap.dedent(cmd) + "\n",
    )
    c = re.findall(r"\S+", cmd.replace("\\\n", ""))
    if not cfg.args.dry_run:
        if shell:
            log.debug("Running command in shell.")
            code = subprocess.call(cmd, shell=shell)
        else:
            code = subprocess.call(c, shell=shell)
        if code != 0:
            log.error(
                "Process for command below failed with error:\n'%s'\nTerminating pipeline.\n",
                textwrap.dedent(cmd),
            )
            sys.exit(code)
        if not shell:
            usage = resource.getrusage(resource.RUSAGE_SELF)
            log.debug(
                "Maximum used memory so far: {:.2f}Gb".format(
                    usage.ru_maxrss / 1e6
                )
            )
    return code


def prep_demo(overwrite=False) -> None:
    download_test_data(cfg.args.dirs["input"][0], overwrite)
    download_test_model(cfg.args.dirs["ilastik"], overwrite)


def download_test_data(output_dir, overwrite=False) -> None:
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    urls = [
        ("imcpipeline-example_pannel.csv", DEMO_CSV),
        ("imcpipeline-demo_data.zip", DEMO_DATA),
    ]

    for fln, url in urls:
        fln = pjoin(output_dir, fln)
        if os.path.exists(fln) and not overwrite:
            log.debug("File exists, skipping: '%s'", url)
            continue
        log.info("Downloading test file: '%s'", url)
        if not os.path.exists(fln):
            request.urlretrieve(url, fln)


def download_test_model(output_dir, overwrite=False) -> None:
    os.makedirs(os.path.abspath(output_dir), exist_ok=True)
    urls = [
        ("demo_ilastik_model.zip", "demo.ilp", DEMO_ILASTIK_MODEL),
    ]
    for zipfln, fln, url in urls:
        zipfln = pjoin(output_dir, zipfln)
        fln = pjoin(output_dir, fln)
        if os.path.exists(fln) and not overwrite:
            log.debug("File exists, skipping: '%s'", url)
            continue
        log.info("Downloading test model: '%s'", url)
        request.urlretrieve(url, zipfln)
        log.info("Unzipping test model: '%s'", url)
        run_shell_command(f"unzip -d {output_dir} {zipfln}")



#!/usr/bin/env python

"""
A temporary script to convert the outputs of the first pipeline to the current
version.
"""

import os
import re

import pandas as pd

from imc.types import Path


df = pd.read_csv('metadata/annotation.csv').set_index('acquisition_date')
df.index = df.index.astype(str)

cols = ['roi_number', 'sample_number', 'panorama_number', 'roi_names']
# get roi_number to be the second label
rois = (
        df[cols[0]]
    .str.split(',')
    .apply(pd.Series).stack()
    .reset_index(level=1, drop=True)
    .rename(cols[0]))

rescaled = list()
for col in cols[1:]:
    rescaled.append(
        df[col]
        .str.split(',')
        .apply(pd.Series)
        .stack()
        .reset_index(level=1, drop=True)
        .rename(col)
        .to_frame()
        .set_index(rois, append=True)
        .squeeze())

df2 = pd.DataFrame(rescaled).T.reset_index(level=1)
df2 = df2.join(df.drop(cols, axis=1))
df2['roi_name'] = df2.index.astype(str) + "-" + df2['roi_number']

df2 = df2.set_index(['roi_name'], append=True).reorder_levels([1, 0], axis=0)
df2.to_csv('metadata/annotation.per_roi.csv')


# Rename files to use only the sample/roi name
for sample, row in df.dropna(subset=['acquisition_name']).iterrows():
    p = Path("processed") / sample / "tiffs"
    if not p.is_dir():
        continue
    files = pd.Series([q.parts[-1] for q in p.iterdir()])
    files = files[files.str.contains(row['acquisition_name'])].sort_values()
    ex = files.str.extract(r"^(" + row['acquisition_name'] + r")_.*_r(\d+)_.*?_ac(_\w+.\w+)")
    repl = sample + '-' + ex[1].astype(str) + ex[2]
    for f, t in zip(files.tolist(), repl):
        print(p / f, p / t)
        os.rename(p / f, p / t)


for sample in df['sample_name'].unique():
    # Rename file endings
    cmds = [
        f"rename 's/_ilastik_s2_Probabilities_mask.tiff/_full_mask.tiff/g' {p}/*.tiff",
        f"rename 's/_ilastik_s2_Probabilities_NucMask.tiff/_full_nucmask.tiff/g' {p}/*.tiff",
        f"rename 's/_ilastik_s2_Probabilities.tiff/_Probabilities.tiff/g' {p}/*.tiff"]
    for cmd in cmds:
        os.system(cmd)

# rename ROI endings
df = pd.read_csv('metadata/annotation.csv')
pat = re.compile(r'(.*)_s\d+_p\d+_r(\d+)_a\d+_ac_(.*)')
for _, row in df.query('toggle').iterrows():
    p = Path("processed") / row['sample_name'] / "tiffs"
    files = list(p.glob(f'*_r{row["roi_number"]}_*'))
    print(files)
    for file in files:
        m = re.match(pat, str(file))
        if m:
            _pre, roi_n, ext = m.groups()
            roi_n = roi_n.zfill(2)
            pre = Path(_pre)
            new = pre.parent / (pre.parts[-1].replace(row['acquisition_name'], row['roi_name']) + "_" + ext)
            print(file, new)
            file.replace(new)


for sample in df['sample_name'].unique():
    # rename ometiff folder (or better find a way to have channel metadata more accessible)
    p = (Path("processed") / sample / "ometiff").absolute()
    for d in p.iterdir():
        print(d, p / sample)
        os.rename(d, p / sample)
        d = p / sample

        # and files (not touching the ROI names for now)
        for f in d.iterdir():
            print(f, str(f).replace(row.acquisition_name, sample))
            os.rename(f, str(f).replace(row.acquisition_name, sample))



#!/usr/bin/env

"""
"""

import os
from os.path import join as pjoin
import logging

try:
    # Even though there is no "imcpipeline/_version" file,
    # it should be generated by
    # setuptools_scm when building the package
    from imcpipeline._version import __version__
except ImportError:
    from setuptools_scm import get_version as _get_version

    __version__ = _get_version(root="..", relative_to=__file__)


def setup_logging(level=logging.INFO):
    # create logger with 'imcpipeline'
    logger = logging.getLogger("imcpipeline")
    logger.setLevel(level)
    # create file handler which logs even debug messages
    fh = logging.FileHandler(pjoin(os.path.expanduser("~"), ".imcpipeline.log"))
    fh.setLevel(logging.DEBUG)
    # create console handler
    ch = logging.StreamHandler()
    ch.setLevel(level)
    # create formatter and add it to the handlers
    fmt = "imcpipeline.v{}:%(module)s:L%(lineno)d ".format(__version__)
    fmt += "(%(funcName)s) [%(levelname)s] %(asctime)s > %(message)s"
    formatter = logging.Formatter(fmt, datefmt="%Y-%m-%d %H:%M:%S")
    fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(fh)
    logger.addHandler(ch)
    return logger


LOGGER = setup_logging()

DOCKER_IMAGE = "afrendeiro/cellprofiler"
DEMO_ILASTIK_MODEL = "https://wcm.box.com/shared/static/4rlm4ty6e0uaf35cp7sm279w3anyoz76.zip"
DEMO_DATA = "https://wcm.box.com/shared/static/wdbntdtsq95dfyffodiz7kb5bmdj5x5u"
DEMO_CSV = "https://wcm.box.com/shared/static/yy24b47sb730twi2zlmo3y4qvi7kc6xd"



#!/usr/bin/env python

"""
A pipeline for the processing of imaging mass cytometry data.
"""

import os
from glob import glob
from os.path import join as pjoin
import sys
import argparse
import shutil
import re
import tempfile

import pandas as pd  # type: ignore
from imctools.scripts import ometiff2analysis  # type: ignore
from imctools.scripts import ome2micat
from imctools.scripts import probablity2uncertainty
from imctools.scripts import convertfolder2imcfolder
from imctools.scripts import exportacquisitioncsv

from imcpipeline import config as cfg  # type: ignore
from imcpipeline import LOGGER as log, DOCKER_IMAGE
from imcpipeline.utils import (
    run_shell_command,
    prep_demo,
    check_ilastik,
    check_requirements,
    docker_or_singularity,
)


STEPS = [
    "prepare",
    "train",
    "predict",
    "segment",
    "quantify",
    "uncertainty",
]

STEPS_INDEX = dict(enumerate(STEPS))

#  DIRS = ['base', 'input', 'analysis', 'ilastik', 'ome', 'cp', 'histocat' 'uncertainty']


def main() -> int:
    log.info("Starting pipeline")
    cfg.args = get_cli_arguments()

    if cfg.args.demo:
        log.info("Running demo data. Output will be in '%s'.", cfg.args.dirs["input"][0])
        prep_demo()

    for name, path in cfg.args.dirs.items():
        if name not in ["input"]:
            os.makedirs(path, exist_ok=True)

    try:
        cfg.args.step = STEPS_INDEX[int(cfg.args.step)]
    except (ValueError, IndexError):
        pass
    finally:
        if cfg.args.step == "all":
            for step in STEPS[:-1]:
                log.info("Doing '%s' step.", step)
                code = globals()[step]()
                log.info("Done with '%s' step.", step)
        else:
            for step in cfg.args.step.split(","):
                print(step)
                log.info("Doing '%s' step.", step)
                code = globals()[step]()
                log.info("Done with '%s' step.", step)
    log.info("Pipeline run completed!")
    return code


def get_cli_arguments() -> argparse.Namespace:
    epilog = (
        "Read all the documentation at: https://imcpipeline.readthedocs.io\n"
        "Issues/feature requests at: https://github.com/elementolab/imcpipeline"
    )

    parser = argparse.ArgumentParser(prog="imcpipeline", epilog=epilog)

    # Demo mode
    parser.add_argument("--demo", dest="demo", action="store_true")

    # Software
    choices = ["docker", "singularity"]
    msg = (
        f"The container engine to use, one of {''.join(choices)}."
        "Default is not using any but CellProfiler must be in PATH."
    )
    parser.add_argument(
        "--container", dest="containerized", default=None, choices=choices, help=msg,
    )
    msg = f"The container image to use. Defaults to {DOCKER_IMAGE}"
    parser.add_argument("--image", dest="container_image", default=DOCKER_IMAGE, help=msg)

    msg = "Location to store external libraries if needed."
    parser.add_argument(
        "--external-lib-dir", dest="external_lib_dir", default=pjoin("lib", "external"), help=msg,
    )
    # # if cellprofiler locations do not exist, clone to some default location
    msg = (
        "In case a containerized version of cellprofiler is not used, "
        "this is a string with the executable for cellprofiler. "
        "Supports UNIX pipes and bash builtins - use with caution. "
        "This could be used for example to activate an environment prior "
        "to execution. Defaults to 'cellprofiler'."
    )
    parser.add_argument(
        "--cellprofiler-exec", dest="cellprofiler_exec", default=None, help=msg,
    )
    msg = "Path to CellProfiler pipeline. If not given will be cloned."
    parser.add_argument(
        "--cellprofiler-pipeline-path",
        dest="cellprofiler_pipeline_path",
        default=None,  # "lib/external/ImcSegmentationPipeline/"
        help=msg,
        type=os.path.abspath,
    )
    msg = "Path to CellProfiler plugins. If not given will be cloned."
    parser.add_argument(
        "--cellprofiler-plugin-path",
        dest="cellprofiler_plugin_path",
        default=None,  # "lib/external/ImcPluginsCP"
        help=msg,
    )
    msg = "Path to Ilastik. If not given will be downloaded."
    parser.add_argument(
        "--ilastik-path",
        dest="ilastik_sh_path",
        default=None,  # "lib/external/ilastik-1.3.3post2-Linux/run_ilastik.sh",
        help=msg,
    )
    # Input
    parser.add_argument("--file-regexp", dest="file_regexp", default=".*.zip")
    msg = "Path to CSV annotation with pannel information."
    parser.add_argument("--csv-pannel", dest="csv_pannel", default=None, help=msg)
    msg = "Column in CSV with metal tag annotation."
    parser.add_argument(
        "--csv-pannel-metal", dest="csv_pannel_metal", default="Metal Tag", help=msg,
    )
    msg = "Column in CSV with boolean annotation of whether the channel be used for ilastik model."
    parser.add_argument(
        "--csv-pannel-ilastik", dest="csv_pannel_ilastik", default="ilastik", help=msg,
    )
    parser.add_argument("--csv-pannel-full", dest="csv_pannel_full", default="full")
    msg = (
        "Directory with input files. More than one value is possible "
        " -just make sure this option is not immediately before the positional argument."
    )
    parser.add_argument(
        "-i", "--input-dirs", nargs="+", dest="input_dirs", default=None, help=msg,
    )

    # Pre-trained model for classification (ilastik)
    msg = "Path to pre-trained ilastik model. If not given will start interactive session."
    parser.add_argument("-m", "--ilastik-model", dest="ilastik_model", default=None, help=msg)

    msg = "Whether existing files should be overwritten."
    parser.add_argument("--overwrite", dest="overwrite", action="store_true")

    # Pipeline steps
    choices = STEPS + [str(x) for x in range(len(STEPS))]
    msg = (
        "Step of the pipeline to perform. 'all' will perform all in sequence."
        f"Options: {', '.join(choices)}"
    )
    parser.add_argument("-s", "--step", dest="step", default="all", choices=choices, help=msg)
    msg = "Whether to not actually run any command. Useful for testing."
    parser.add_argument("-d", "--dry-run", dest="dry_run", action="store_true", help=msg)
    msg = "Directory for output files."
    parser.add_argument("-o", "--output-dir", dest="output_dir", default=None, help=msg)

    # Parse and complete with derived info
    args = parser.parse_args()
    if not args.demo and args.output_dir is None:
        parser.error("'-o/--output-dir' must be given if not in demo mode.")
    if args.demo:
        args.input_dirs = [os.path.abspath("imcpipeline_demo_data")]
        args.output_dir = "imcpipeline_demo_data"
        args.csv_pannel = pjoin("imcpipeline_demo_data", "imcpipeline-example_pannel.csv")
        args.ilastik_model = pjoin("imcpipeline_demo_data", "ilastik", "demo.ilp")
        args.step = "all"
    dirs = dict()
    args.output_dir = os.path.abspath(args.output_dir)
    dirs["base"] = args.output_dir
    dirs["input"] = args.input_dirs or [pjoin(args.output_dir, "input_data")]
    dirs["analysis"] = pjoin(dirs["base"], "tiffs")
    dirs["ilastik"] = pjoin(dirs["base"], "ilastik")
    dirs["ome"] = pjoin(dirs["base"], "ometiff")
    dirs["cp"] = pjoin(dirs["base"], "cpout")
    dirs["histocat"] = pjoin(dirs["base"], "histocat")
    dirs["uncertainty"] = pjoin(dirs["base"], "uncertainty")
    args.dirs = dirs

    dirbind = {"docker": "-v", "singularity": "-B"}
    if args.containerized is not None:
        args.dirbind = dirbind[args.containerized]
    elif args.containerized is None and args.cellprofiler_exec is not None:
        pass
    elif args.containerized is None and args.cellprofiler_exec is None:
        if shutil.which("cellprofiler"):
            args.cellprofiler_exec = "cellprofiler"
        else:
            log.info(
                "Neither a container engine was given nor a cellprofiler executable was found."
            )
            try:
                args.containerized = docker_or_singularity()
                args.dirbind = dirbind[args.containerized]
            except ValueError:
                parser.error("Neither docker, singularity or a cellprofiler executable were found!")
            log.info("Found '%s', will use that.", args.containerized)

    if args.csv_pannel is None:
        if args.input_dirs is not None:
            args.csv_pannel = pjoin(args.input_dirs[0], "imcpipeline-example_pannel.csv")
    args.parsed_csv_pannel = pjoin(args.dirs["base"], "pannel_data.acquired_channels.csv")

    args.suffix_mask = "_mask.tiff"
    args.suffix_probablities = "_Probabilities"
    args.list_analysis_stacks = [
        (args.csv_pannel_ilastik, "_ilastik", 1),
        (args.csv_pannel_full, "_full", 0),
    ]

    return args


def prepare() -> int:
    """
    Extract MCD files and prepare input for ilastik.
    """

    def export_acquisition() -> None:
        re_fn = re.compile(cfg.args.file_regexp)

        for fol in cfg.args.dirs["input"]:
            for fln in os.listdir(fol):
                if re_fn.match(fln):
                    fn_full = pjoin(fol, fln)
                    log.info("Extracting MCD file '%s'.", fn_full)
                    if cfg.args.dry_run:
                        continue
                    convertfolder2imcfolder.convert_folder2imcfolder(
                        fn_full, out_folder=cfg.args.dirs["ome"], dozip=False
                    )
        if cfg.args.dry_run:
            return
        exportacquisitioncsv.export_acquisition_csv(
            cfg.args.dirs["ome"], fol_out=cfg.args.dirs["cp"]
        )

    def join_pannel_with_acquired_channels(directory=None) -> None:
        to_replace = [
            ("-", ""),
            ("_", ""),
            (" ", ""),
        ]
        # read pannel
        pannel = pd.read_csv(cfg.args.csv_pannel, index_col=0)
        # read acquisition metadata
        if directory is None:
            pattern = pjoin(cfg.args.dirs["ome"], "*", "*_AcquisitionChannel_meta.csv")
        else:
            pattern = pjoin(directory, "*_AcquisitionChannel_meta.csv")
        metas = glob(pattern)
        if not metas:
            raise ValueError(f"No '{pattern}' files  found!")
        if len(metas) != 1:
            raise ValueError(f"More than one '{pattern}' files found!")

        acquired = pd.read_csv(metas[0])
        acquired = acquired[["ChannelLabel", "ChannelName", "OrderNumber"]]

        # remove parenthesis from metal column
        acquired["ChannelName"] = acquired["ChannelName"].str.replace("(", "").str.replace(")", "")
        # clean up the channel name
        for __k, __v in to_replace:
            acquired["ChannelLabel"] = acquired["ChannelLabel"].str.replace(__k, __v)
        acquired["ChannelLabel"] = acquired["ChannelLabel"].fillna("<EMPTY>")
        acquired = acquired.loc[
            ~acquired["ChannelLabel"].isin(["X", "Y", "Z"]), :
        ].drop_duplicates()
        acquired.index = acquired["ChannelLabel"] + "(" + acquired["ChannelName"] + ")"

        # Check matches, report missing
        __c = acquired.index.isin(pannel.index)
        if not __c.all():
            miss = "\n - ".join(acquired.loc[~__c, "ChannelLabel"])
            raise ValueError(
                f"Given reference pannel '{cfg.args.csv_pannel}'"
                f" is missing the following channels: \n - {miss}"
            )

        # align and sort by acquisition
        joint_pannel = acquired.join(pannel).sort_values("OrderNumber")

        # make sure order of ilastik channels is same as the original pannel
        # this important in order for the channels to always be the same
        # and the ilastik models to be reusable
        assert all(
            pannel.query("ilastik == True").index == joint_pannel.query("ilastik == True").index
        )

        # If all is fine, save annotation with acquired channels and their order
        joint_pannel.to_csv(cfg.args.parsed_csv_pannel, index=True)

    def prepare_histocat() -> None:
        if not os.path.exists(cfg.args.dirs["histocat"]):
            os.makedirs(cfg.args.dirs["histocat"])
        for fol in os.listdir(cfg.args.dirs["ome"]):
            if cfg.args.dry_run:
                continue
            ome2micat.omefolder2micatfolder(
                pjoin(cfg.args.dirs["ome"], fol), cfg.args.dirs["histocat"], dtype="uint16",
            )

        pannel = (
            cfg.args.parsed_csv_pannel
            if os.path.exists(cfg.args.parsed_csv_pannel)
            else cfg.args.csv_pannel
        )

        for fol in os.listdir(cfg.args.dirs["ome"]):
            sub_fol = pjoin(cfg.args.dirs["ome"], fol)
            for img in os.listdir(sub_fol):
                if not img.endswith(".ome.tiff"):
                    continue
                basename = img.rstrip(".ome.tiff")
                log.info("Preparing OME-tiff directory '%s'.", img)
                for (col, suffix, addsum) in cfg.args.list_analysis_stacks:
                    if cfg.args.dry_run:
                        continue
                    ometiff2analysis.ometiff_2_analysis(
                        pjoin(sub_fol, img),
                        cfg.args.dirs["analysis"],
                        basename + suffix,
                        pannelcsv=pannel,
                        metalcolumn=cfg.args.csv_pannel_metal,
                        usedcolumn=col,
                        addsum=addsum,
                        bigtiff=False,
                        pixeltype="uint16",
                    )

    @check_requirements
    def prepare_ilastik() -> None:
        if cfg.args.containerized:
            extra = (
                "--name cellprofiler_prepare_ilastik --rm"
                if cfg.args.containerized == "docker"
                else ""
            )
            cmd = f"""
        {cfg.args.containerized} run \\
        {extra} \\
            {cfg.args.dirbind} {cfg.args.dirs['base']}:/data:rw \\
            {cfg.args.dirbind} {cfg.args.cellprofiler_plugin_path}:/ImcPluginsCP:ro \\
            {cfg.args.dirbind} {cfg.args.cellprofiler_pipeline_path}:/ImcSegmentationPipeline:ro \\
            {cfg.args.container_image} \\
                --run-headless --run \\
                --plugins-directory /ImcPluginsCP/plugins/ \\
                --pipeline /ImcSegmentationPipeline/cp3_pipelines/1_prepare_ilastik.cppipe \\
                -i /{cfg.args.dirs['analysis'].replace(cfg.args.dirs['base'], 'data')}/ \\
                -o /{cfg.args.dirs['ilastik'].replace(cfg.args.dirs['base'], 'data')}/"""
        else:
            cmd = f"""
            {cfg.args.cellprofiler_exec} \\
                --run-headless --run \\
                --plugins-directory {cfg.args.cellprofiler_plugin_path}/plugins/ \\
                --pipeline {cfg.args.cellprofiler_pipeline_path}/cp3_pipelines/1_prepare_ilastik.cppipe \\
                -i {cfg.args.dirs['analysis']}/ \\
                -o {cfg.args.dirs['ilastik']}/"""

        # {cfg.args.dirbind} /tmp/.X11-unix:/tmp/.X11-unix:ro \\
        # -e DISPLAY=$DISPLAY \\
        run_shell_command(cmd)

    def fix_spaces_in_folders_files(directory):
        for path, folders, files in os.walk(directory):
            for f in files:
                os.rename(
                    pjoin(path, f), pjoin(path, f.replace(" ", "_")),
                )
            for i, _ in enumerate(folders):
                new_name = folders[i].replace(" ", "_")
                os.rename(pjoin(path, folders[i]), pjoin(path, new_name))
                folders[i] = new_name

    e = os.path.exists(pjoin(cfg.args.dirs["cp"], "acquisition_metadata.csv"))
    if cfg.args.overwrite or (not cfg.args.overwrite and not e):
        log.info("Expanding directories from MCD files.")
        export_acquisition()
    else:
        log.info("Overwrite is false and files exist. Skipping export from MCD.")

    e = len(glob(pjoin(cfg.args.dirs["analysis"], "*_full.tiff"))) > 0
    if cfg.args.overwrite or (not cfg.args.overwrite and not e):
        if not cfg.args.dry_run:
            try:
                join_pannel_with_acquired_channels()
            except ValueError:
                log.error("Failed formatting channel names with provided pannel CSV metadata.")
        prepare_histocat()
    else:
        log.info("Overwrite is false and files exist. Skipping conversion to OME-tiff.")
    e = len(glob(pjoin(cfg.args.dirs["ilastik"], "*_w500_h500.h5"))) > 0
    if cfg.args.overwrite or (not cfg.args.overwrite and not e):
        prepare_ilastik()
    else:
        log.info("Overwrite is false and files exist. Skipping preparing ilastik files.")

    fix_spaces_in_folders_files(cfg.args.dirs["base"])
    return 0


@check_ilastik
def train() -> int:
    """
    Train an ilastik pixel classification model.

    Inputs are the files in ilastik/*.h5
    """
    if cfg.args.step == "all" and cfg.args.ilastik_model is not None:
        log.info("Pre-trained model provided. Skipping training step.")
        return 0
    else:
        log.info("No model provided. Launching interactive ilastik session.")
        cmd = f"""{cfg.args.ilastik_sh_path}"""
        return run_shell_command(cmd)


@check_ilastik
def predict() -> int:
    """
    Use a trained ilastik model to classify pixels in an IMC image.
    """
    # Check if step should be skipped:
    # for each "_s2.h5" file there is a "_s2_Probabilities.tiff".
    inputs = glob(f"{cfg.args.dirs['analysis']}/*_s2.h5")
    exists = [os.path.exists(f.replace("_s2.h5", "_s2_Probabilities.tiff")) for f in inputs]
    if all(exists) and not cfg.args.overwrite:
        log.info("All output predictions exist. Skipping prediction step.")
        return 0

    cmd = f"""{cfg.args.ilastik_sh_path} \\
        --headless \\
        --readonly \\
        --export_source probabilities \\
        --project {cfg.args.ilastik_model} \\
        """
    # Shell expansion of input files won't happen in subprocess call
    cmd += " ".join([x.replace(" ", r"\ ") for x in inputs])
    return run_shell_command(cmd)


@check_requirements
def segment() -> int:
    """
    Segment a TIFF with class probabilities into nuclei and cells using cellprofiler.
    """
    # Check if step should be skipped:
    # for each "_s2_Probabilities.tiff" file there is a "_s2_Probabilities_mask.tiff".
    exists = [
        os.path.exists(f.replace("_s2_Probabilities.tiff", "_s2_Probabilities_mask.tiff"))
        for f in glob(f"{cfg.args.dirs['analysis']}/*_s2_Probabilities.tiff")
    ]
    if all(exists) and not cfg.args.overwrite:
        log.info("Segmentations already exist for all images. Skipping segment step.")
        return 0

    extra = "--name cellprofiler_segment --rm" if cfg.args.containerized == "docker" else ""

    if cfg.args.containerized:
        cmd = f"""{cfg.args.containerized} run \\
        {extra} \\
        {cfg.args.dirbind} {cfg.args.dirs['base']}:/data:rw \\
        {cfg.args.dirbind} {cfg.args.cellprofiler_plugin_path}:/ImcPluginsCP:ro \\
        {cfg.args.dirbind} {cfg.args.cellprofiler_pipeline_path}:/ImcSegmentationPipeline:ro \\
        {cfg.args.container_image} \\
            --run-headless --run \\
            --plugins-directory /ImcPluginsCP/plugins/ \\
            --pipeline /ImcSegmentationPipeline/cp3_pipelines/2_segment_ilastik.cppipe \\
            -i /{cfg.args.dirs['analysis'].replace(cfg.args.dirs['base'], 'data')}/ \\
            -o /{cfg.args.dirs['analysis'].replace(cfg.args.dirs['base'], 'data')}/"""
    else:
        cmd = f"""
        {cfg.args.cellprofiler_exec} \\
            --run-headless --run \\
            --plugins-directory {cfg.args.cellprofiler_plugin_path}/plugins/ \\
            --pipeline {cfg.args.cellprofiler_pipeline_path}/cp3_pipelines/2_segment_ilastik.cppipe \\
            -i {cfg.args.dirs['analysis']}/ \\
            -o {cfg.args.dirs['analysis']}/"""
    return run_shell_command(cmd)


@check_requirements
def quantify() -> int:
    """
    Quantify the intensity of each channel in each single cell.
    """
    # Check if step should be skipped:
    exists = os.path.exists(pjoin(cfg.args.dirs["cp"], "cell.csv"))
    if exists and not cfg.args.overwrite:
        log.info("Quantifications already exist. Skipping quantify step.")
        return 0

    # For this step, the number of channels should be updated
    # in the pipeline file (line 126 and 137).
    pipeline_file = pjoin(
        cfg.args.cellprofiler_pipeline_path, "cp3_pipelines", "3_measure_mask_basic.cppipe",
    )
    new_pipeline_file = tempfile.NamedTemporaryFile(dir=".", suffix=".cppipe").name

    # Update channel number with pannel for quantification step
    pannel = (
        cfg.args.parsed_csv_pannel
        if os.path.exists(cfg.args.parsed_csv_pannel)
        else cfg.args.csv_pannel
    )
    cfg.args.channel_number = pd.read_csv(pannel).query("full == 1").shape[0]

    default_channel_number = r"\xff\xfe2\x004\x00"
    new_channel_number = (
        str(str(cfg.args.channel_number).encode("utf-16")).replace("b'", "").replace("'", "")
    )
    log.info("Changing the channel number to %i.", cfg.args.channel_number)

    with open(pipeline_file, "r") as ihandle:
        with open(new_pipeline_file, "w") as ohandle:
            c = ihandle.read()
            cc = c.replace(default_channel_number, new_channel_number)
            ohandle.write(cc)

    if cfg.args.containerized:
        extra = "--name cellprofiler_quantify --rm" if cfg.args.containerized == "docker" else ""
        cmd = f"""{cfg.args.containerized} run \\
        {extra} \\
        {cfg.args.dirbind} {cfg.args.dirs['base']}:/data:rw \\
        {cfg.args.dirbind} {cfg.args.cellprofiler_plugin_path}:/ImcPluginsCP:ro \\
        {cfg.args.dirbind} {os.path.abspath(".")}:/ImcSegmentationPipeline:ro \\
        {cfg.args.container_image} \\
            --run-headless --run \\
            --plugins-directory /ImcPluginsCP/plugins/ \\
            --pipeline /ImcSegmentationPipeline/{os.path.basename(new_pipeline_file)} \\
            -i /{cfg.args.dirs['analysis'].replace(cfg.args.dirs['base'], 'data')}/ \\
            -o /{cfg.args.dirs['cp'].replace(cfg.args.dirs['base'], 'data')}"""
    else:
        cmd = f"""
        {cfg.args.cellprofiler_exec} \\
            --run-headless --run \\
            --plugins-directory {cfg.args.cellprofiler_plugin_path}/plugins/ \\
            --pipeline {new_pipeline_file} \\
            -i {cfg.args.dirs['analysis']}/ \\
            -o {cfg.args.dirs['cp']}"""

    code = run_shell_command(cmd)

    os.remove(new_pipeline_file)
    return code


def uncertainty() -> int:
    """
    Produce maps of model uncertainty.

    This step requires LZW decompression which is given by the `imagecodecs`
    Python library (has extensive system-level dependencies in Ubuntu).
    """

    # Check if step should be skipped:
    # for each "tiffs/*_s2_Probabilities.tiff" file there is
    # a "uncertainty/*_s2_Probabilities_uncertainty.tiff".
    exists = [
        os.path.exists(
            pjoin(
                cfg.args.dirs["uncertainty"],
                os.path.basename(f).replace(".tiff", "_uncertainty.tiff"),
            )
        )
        for f in glob(f"{cfg.args.dirs['analysis']}/*_s2_Probabilities.tiff")
    ]
    if all(exists) and not cfg.args.overwrite:
        log.info("Segmentations already exist for all images. Skipping segment step.")
        return 0

    for fn in os.listdir(cfg.args.dirs["ilastik"]):
        if fn.endswith(cfg.args.suffix_probablities + ".tiff"):
            log.info("Exporting uncertainties for image '%s'.", fn)
            probablity2uncertainty.probability2uncertainty(
                pjoin(cfg.args.dirs["ilastik"], fn), cfg.args.dirs["uncertainty"]
            )

    for fn in os.listdir(cfg.args.dirs["analysis"]):
        if fn.endswith(cfg.args.suffix_probablities + ".tiff"):
            log.info("Exporting uncertainties for image '%s'.", fn)
            probablity2uncertainty.probability2uncertainty(
                pjoin(cfg.args.dirs["analysis"], fn), cfg.args.dirs["uncertainty"]
            )

    for fol in os.listdir(cfg.args.dirs["ome"]):
        ome2micat.omefolder2micatfolder(
            pjoin(cfg.args.dirs["ome"], fol),
            cfg.args.dirs["histocat"],
            fol_masks=cfg.args.dirs["analysis"],
            mask_suffix=cfg.args.suffix_mask,
            dtype="uint16",
        )
    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        sys.exit(1)



#! /usr/bin/env python

"""
Installer script for the ``imc`` library and the ``imcpipeline`` pipeline.

Install with ``pip install .``.
"""

from setuptools import setup, find_packages  # type: ignore


def parse_requirements(req_file):
    """Parse requirements.txt files."""
    reqs = open(req_file).read().strip().split("\n")
    reqs = [r for r in reqs if not r.startswith("#")]
    return [r for r in reqs if "#egg=" not in r]


REQUIREMENTS_FILE = "requirements.txt"
DEV_REQUIREMENTS_FILE = "requirements.dev.txt"
README_FILE = "README.md"

# Requirements
requirements = parse_requirements(REQUIREMENTS_FILE)
requirements_dev = parse_requirements(DEV_REQUIREMENTS_FILE)

# Description
long_description = open(README_FILE).read()


# setup
setup(
    name="imcpipeline",
    packages=find_packages(),
    use_scm_version={
        "write_to": "imcpipeline/_version.py",
        "write_to_template": '__version__ = "{version}"\n',
    },
    entry_points={
        "console_scripts": [
            "imcrunner = imcpipeline.runner:main",
            "imcpipeline = imcpipeline.pipeline:main",
        ]
    },
    description="A pipeline and utils for IMC data analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Typing :: Typed",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords=",".join(
        [
            "computational biology",
            "bioinformatics",
            "imaging",
            "mass cytometry",
            "mass spectrometry",
        ]
    ),
    url="https://github.com/elementolab/imcpipeline",
    project_urls={
        "Bug Tracker": "https://github.com/elementolab/imcpipeline/issues",
        "Documentation": "https://github.com/elementolab/imcpipeline/README.md",
        "Source Code": "https://github.com/elementolab/imcpipeline",
    },
    author=u"Andre Rendeiro",
    author_email="andre.rendeiro@pm.me",
    license="GPL3",
    setup_requires=["setuptools_scm"],
    install_requires=requirements,
    tests_require=requirements_dev,
    extras_require={"dev": requirements_dev},
    package_data={"imcpipeline": ["logo.png", "docs.md"]},
)
